-- Common Lua functions to help process strings

StringUtils = {}

-- Return true or false indicating if the input string starts with the second input string
StringUtils.StartsWith = function (string1, string2)
	return string.sub(string1, 1, string.len(string2)) == string2
end

-- Return the string without preceeding or trailing whitespace
StringUtils.Trim = function (input)
    if not input or type(input) ~= "string" then return nil end
	return (input:gsub("^%s*(.-)%s*$", "%1"))
end

-- Return a flag indicating if the input string is empty
-- Returns true if the input value is nil, not a string, or is empty
-- Returns false if the input is a string with contents that are not whitespace
StringUtils.IsEmptyOrNil = function (input)
	if input == nil or type(input) ~= "string" then return true end
	return input:match("^%s*$")
end

-- Pads the input string if length is shorter than the input length
-- By default, pads the right side; pads left side if padleft = true
StringUtils.Pad = function (input, length, padleft)
	local inputLength = #input
    if inputLength >= length then
        return input
    end

    local padding = string.rep(" ", length - inputLength)
    if padleft then
        return padding .. input
	end
	return input .. padding
end

-- Center a string
StringUtils.Center = function (input, totallength)
    local inputLength = #input
    if inputLength >= totallength then
        return input
    end
    
    local padding = totallength - inputLength
    local leftPadding = math.floor(padding / 2)
    local rightPadding = padding - leftPadding
    
    return string.rep(" ", leftPadding) .. input .. string.rep(" ", rightPadding)
end